import functools
from typing import Callable, Any, Iterator, Optional

import wrapt
from opentelemetry import trace as otel_trace, context
from opentelemetry.trace import StatusCode
from opentelemetry.trace.span import Span
import logging

logger = logging.getLogger(__name__)


def start_trace(
    name: Optional[str] = "default",
    attributes: Optional[dict] = None,
    current_context: Optional[otel_trace.SpanContext] = None,
) -> Optional[object]:
    tracer = otel_trace.get_tracer(__name__)

    # Create new span
    span = tracer.start_span(name or "default")
    if attributes:
        span.set_attributes(attributes)

    ctx = otel_trace.set_span_in_context(span, current_context)
    return context.attach(ctx)


def end_trace(
    token: object,
) -> Optional[Span]:
    span = otel_trace.get_current_span()
    
    # End the span and remove span from context
    context.detach(token)
    span.end()

    return span


def trace(
    wrapped: Optional[Callable] = None,
    *,
    name: Optional[str] = None,
    ignore_args: Optional[bool] = None,
    ignore_kwargs: Optional[bool] = None,
    attributes: Optional[dict] = None,
) -> Callable:
    """Decorator to trace a function with OpenTelemetry.

    Args:
        name (Optional[str]): Optional span name. Defaults to the wrapped function's name.
        ignore_args (Optional[bool]): Whether to ignore positional arguments, tracking is one default.
        ignore_kwargs (Optional[bool]): Whether to ignore keyword arguments, tracking is on by default.
        attributes (Optional[Dict[str, Any]]): Additional attributes to add to the span.
    """
    if wrapped is None:
        return functools.partial(
            trace,
            name=name,
            ignore_args=ignore_args,
            ignore_kwargs=ignore_kwargs,
            attributes=attributes,
        )

    @wrapt.decorator
    def wrapper(wrapped: Callable, instance, args, kwargs) -> Any:
        tracer = otel_trace.get_tracer(__name__)
        with tracer.start_as_current_span(name or wrapped.__name__) as span:
            try:
                if attributes:
                    span.set_attributes(attributes)

                if args and not ignore_args:
                    if len(args) == 1:
                        span.set_attribute("arg", args[0])
                    elif len(args) > 1:
                        span.set_attribute("args", args)
                if kwargs and not ignore_kwargs:
                    span.set_attributes(kwargs)

                ret = wrapped(*args, **kwargs)

                span.set_status(StatusCode.OK)
                return ret
            except Exception as e:
                span.set_status(StatusCode.ERROR, str(e))
                span.record_exception(e)
                raise e

    return wrapper(wrapped)  # type: ignore


def trace_lambda(
    name: Optional[str] = None,
    ignore_args: Optional[bool] = None,
    attributes: Optional[dict] = None,
) -> Callable:
    """Decorator to trace a lambda function with OpenTelemetry."""
    return trace(name=name, ignore_args=ignore_args, attributes=attributes)


def set_attribute(key: str, value: Any):
    """Set an attribute in the current span."""
    span = otel_trace.get_current_span()
    if span:
        span.set_attribute(key, value)



import logging
import os
import traceback
from contextlib import asynccontextmanager
from importlib.metadata import version

import openai

from .shared.config import Config


def is_openai_v1():
    return version("openai") >= "1.0.0"


def is_azure_openai(instance):
    return is_openai_v1() and isinstance(
        instance._client, (openai.AsyncAzureOpenAI, openai.AzureOpenAI)
    )


def is_metrics_enabled() -> bool:
    return (os.getenv("IUDEX_METRICS_ENABLED") or "true").lower() == "true"


def should_record_stream_token_usage():
    return Config.enrich_token_usage


def _with_image_gen_metric_wrapper(func):
    def _with_metric(duration_histogram, exception_counter):
        def wrapper(wrapped, instance, args, kwargs):
            return func(
                duration_histogram, exception_counter, wrapped, instance, args, kwargs
            )

        return wrapper

    return _with_metric


def _with_embeddings_telemetry_wrapper(func):
    def _with_embeddings_telemetry(
        tracer,
        token_counter,
        vector_size_counter,
        duration_histogram,
        exception_counter,
    ):
        def wrapper(wrapped, instance, args, kwargs):
            return func(
                tracer,
                token_counter,
                vector_size_counter,
                duration_histogram,
                exception_counter,
                wrapped,
                instance,
                args,
                kwargs,
            )

        return wrapper

    return _with_embeddings_telemetry


def _with_chat_telemetry_wrapper(func):
    def _with_chat_telemetry(
        tracer,
        token_counter,
        choice_counter,
        duration_histogram,
        exception_counter,
        streaming_time_to_first_token,
        streaming_time_to_generate,
    ):
        def wrapper(wrapped, instance, args, kwargs):
            return func(
                tracer,
                token_counter,
                choice_counter,
                duration_histogram,
                exception_counter,
                streaming_time_to_first_token,
                streaming_time_to_generate,
                wrapped,
                instance,
                args,
                kwargs,
            )

        return wrapper

    return _with_chat_telemetry


def _with_tracer_wrapper(func):
    def _with_tracer(tracer):
        def wrapper(wrapped, instance, args, kwargs):
            return func(tracer, wrapped, instance, args, kwargs)

        return wrapper

    return _with_tracer


@asynccontextmanager
async def start_as_current_span_async(tracer, *args, **kwargs):
    with tracer.start_as_current_span(*args, **kwargs) as span:
        yield span


def dont_throw(func):
    """
    A decorator that wraps the passed in function and logs exceptions instead of throwing them.

    @param func: The function to wrap
    @return: The wrapper function
    """
    # Obtain a logger specific to the function's module
    logger = logging.getLogger(func.__module__)

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger.debug(
                "OpenLLMetry failed to trace in %s, error: %s",
                func.__name__,
                traceback.format_exc(),
            )
            if Config.exception_logger:
                Config.exception_logger(e)

    return wrapper



# ruff: noqa: E402
# ^ because monkeypatches must import and run before other imports
from .monkeypatches.get_attributes import patched_get_attributes
from .monkeypatches.logging import monkeypatch_LogRecord_getMessage
from .monkeypatches.print import monkeypatch_print

import importlib.util
import logging
import os
import subprocess
from typing import Optional, TypedDict, Union

from opentelemetry._logs import set_logger_provider
from opentelemetry.exporter.otlp.proto.http._log_exporter import OTLPLogExporter
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
from opentelemetry.sdk.environment_variables import (
    OTEL_EXPORTER_OTLP_HEADERS,
    OTEL_EXPORTER_OTLP_LOGS_ENDPOINT,
    OTEL_EXPORTER_OTLP_TIMEOUT,
    OTEL_EXPORTER_OTLP_TRACES_ENDPOINT,
    OTEL_LOG_LEVEL,
    OTEL_SERVICE_NAME,
)
from opentelemetry.sdk.resources import Attributes, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.trace import set_tracer_provider

_logger = logging.getLogger(__name__)

DEFAULT_SERVICE_NAME = "unknown_service:python"
DEFAULT_IUDEX_LOGS_ENDPOINT = "https://api.iudex.ai/resource_logs"
DEFAULT_IUDEX_TRACES_ENDPOINT = "https://api.iudex.ai/traces"
DEFAULT_LOG_LEVEL = logging.INFO
DEFAULT_TIMEOUT = 60

LOG_LEVEL_ATOI = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
    "NOTSET": logging.NOTSET,
}

IUDEX_CONFIGURED = False


class IudexConfig(TypedDict, total=False):
    iudex_api_key: Optional[str]
    service_name: Optional[str]
    instance_id: Optional[str]
    logs_endpoint: Optional[str]
    traces_endpoint: Optional[str]
    log_level: Optional[Union[str, int]]
    git_commit: Optional[str]
    github_url: Optional[str]
    env: Optional[str]
    attributes: Optional[Attributes]
    timeout: Optional[int]
    disable_print: Optional[bool]


class _IudexConfig:
    def __init__(
        self,
        **kwargs,
    ):
        self.iudex_api_key = kwargs.get("iudex_api_key") or os.getenv("IUDEX_API_KEY")
        if not self.iudex_api_key:
            _logger.warning(
                "Missing API key, no telemetry will be sent to Iudex. "
                + "Provide the iudex_api_key parameter or set the IUDEX_API_KEY environment variable."
            )
            return

        self.service_name = (
            kwargs.get("service_name")
            or os.getenv(OTEL_SERVICE_NAME)
            or DEFAULT_SERVICE_NAME
        )
        self.instance_id = kwargs.get("instance_id")

        self.logs_endpoint = (
            kwargs.get("logs_endpoint")
            or os.getenv(OTEL_EXPORTER_OTLP_LOGS_ENDPOINT)
            or DEFAULT_IUDEX_LOGS_ENDPOINT
        )

        self.traces_endpoint = (
            kwargs.get("traces_endpoint")
            or os.getenv(OTEL_EXPORTER_OTLP_TRACES_ENDPOINT)
            or DEFAULT_IUDEX_TRACES_ENDPOINT
        )

        log_level = kwargs.get("log_level") or os.getenv(OTEL_LOG_LEVEL)
        if isinstance(log_level, str):
            log_level = LOG_LEVEL_ATOI.get(log_level.upper())
        self.log_level = log_level or DEFAULT_LOG_LEVEL

        self.git_commit = kwargs.get("git_commit") or os.getenv("GIT_COMMIT")
        if not self.git_commit:
            try:
                self.git_commit = subprocess.check_output(
                    ["git", "rev-parse", "HEAD"]
                ).strip()
            except:
                pass

        self.github_url = kwargs.get("github_url") or os.getenv("GITHUB_URL")

        self.env = kwargs.get("env") or os.getenv("ENV") or os.getenv("ENVIRONMENT")

        self.attributes = kwargs.get("attributes", {})

        self._timeout = kwargs.get("timeout") or int(os.getenv(OTEL_EXPORTER_OTLP_TIMEOUT, DEFAULT_TIMEOUT))

        self.disable_print = kwargs.get("disablePrint") or kwargs.get("disable_print") or False

    def configure(self):
        if not self.iudex_api_key:
            _logger.warning(
                "Missing API key, no telemetry will be sent to Iudex. "
                + "Provide the iudex_api_key parameter or set the IUDEX_API_KEY environment variable."
            )
            return

        global IUDEX_CONFIGURED
        if IUDEX_CONFIGURED:
            return

        # configure common
        attributes = {}
        attributes["service.name"] = self.service_name
        attributes["service.instance.id"] = self.instance_id
        attributes["git.commit"] = self.git_commit
        attributes["github.url"] = self.github_url
        attributes["env"] = self.env
        # clean attributes
        attributes = {
            key: value for key, value in attributes.items() if value is not None
        }

        # add manual attributes
        attributes.update(self.attributes)

        # create resource
        resource = Resource.create(attributes)

        # set default headers to send iudex api key
        headers = {"x-api-key": self.iudex_api_key}
        os.environ[OTEL_EXPORTER_OTLP_HEADERS] = f"x-api-key={self.iudex_api_key}"

        # configure logger
        logger_provider = LoggerProvider(resource=resource)
        log_exporter = OTLPLogExporter(endpoint=self.logs_endpoint, headers=headers, timeout=self._timeout)
        logger_provider.add_log_record_processor(BatchLogRecordProcessor(log_exporter))
        set_logger_provider(logger_provider)
        logging.basicConfig(level=self.log_level)
        # add otel handler to root logger
        configure_logging(log_level=self.log_level)
        if not self.disable_print: 
            # monkeypatch print to emit with otel handler
            monkeypatch_print(LoggingHandler(level=logging.INFO))

        # configure tracer
        trace_provider = TracerProvider(resource=resource)
        span_exporter = OTLPSpanExporter(endpoint=self.traces_endpoint, headers=headers, timeout=self._timeout)
        trace_provider.add_span_processor(BatchSpanProcessor(span_exporter))
        set_tracer_provider(trace_provider)

        IUDEX_CONFIGURED = True


def configure_logging(
    logger_name: Optional[str] = None,
    log_level: Optional[Union[str, int]] = None,
):
    monkeypatch_LogRecord_getMessage()

    logger_handler = LoggingHandler(level=log_level)
    logger_handler._get_attributes = patched_get_attributes

    configure_logger(
        logger_name=logger_name, log_level=log_level, logger_handler=logger_handler
    )
    configure_loguru(log_level=log_level, logger_handler=logger_handler)

def configure_logger(
    logger_name: Optional[str] = None,
    log_level: Optional[Union[str, int]] = None,
    logger_handler: Optional[LoggingHandler] = None,
):
    """Instruments a named logger.

    Useful for non-root loggers with propagate=False.
    This way, the logger still has the instrumented handler to send logs to Iudex.
    """
    if isinstance(log_level, str):
        log_level = LOG_LEVEL_ATOI.get(log_level.upper())
    log_level = log_level or DEFAULT_LOG_LEVEL

    logger = logging.getLogger(logger_name)
    logger.setLevel(log_level)

    if not logger_handler:
        logger_handler = LoggingHandler(level=log_level)
        logger_handler._get_attributes = patched_get_attributes

    logger.addHandler(logger_handler)

    return logger


def configure_loguru(
    log_level: Optional[Union[str, int]] = None,
    logger_handler: Optional[LoggingHandler] = None,
    format: Optional[str] = None,
):
    """Instruments Loguru to send logs to Iudex."""
    if importlib.util.find_spec("loguru") is None:
        return
    from loguru import logger

    if isinstance(log_level, str):
        log_level = LOG_LEVEL_ATOI.get(log_level.upper())
    log_level = log_level or DEFAULT_LOG_LEVEL

    if not logger_handler:
        logger_handler = LoggingHandler(level=log_level)
        logger_handler._get_attributes = patched_get_attributes

    if not format:
        format = "{time:YYYY-MM-DD HH:mm:ss} | {level} | {name} | {message}"

    logger.add(logger_handler, format=format)

    return logger



import json
import logging
import os
import types
from importlib.metadata import version

import openai
from opentelemetry import context as context_api
from opentelemetry.semconv.ai import SpanAttributes

from ..utils import (
    dont_throw,
    is_openai_v1,
    should_record_stream_token_usage,
)

OPENAI_LLM_USAGE_TOKEN_TYPES = ["prompt_tokens", "completion_tokens"]

DEFAULT_MODEL_FOR_ENCODING = "gpt-4"

# tiktoken encodings map for different model, key is model_name, value is tiktoken encoding
tiktoken_encodings = {}

logger = logging.getLogger(__name__)


def should_send_prompts():
    return (
        os.getenv("IUDEX_SHOULD_RECORD_LLM_CONTENT") or "true"
    ).lower() == "true" or context_api.get_value("override_enable_content_tracing")


def _set_span_attribute(span, name, value):
    if value is not None:
        if value != "":
            span.set_attribute(name, value)
    return


def _set_client_attributes(span, instance):
    if not span.is_recording():
        return

    if not is_openai_v1():
        return

    client = instance._client  # pylint: disable=protected-access
    if isinstance(client, (openai.AsyncOpenAI, openai.OpenAI)):
        _set_span_attribute(
            span, SpanAttributes.LLM_OPENAI_API_BASE, str(client.base_url)
        )
    if isinstance(client, (openai.AsyncAzureOpenAI, openai.AzureOpenAI)):
        _set_span_attribute(
            span, SpanAttributes.LLM_OPENAI_API_VERSION, client._api_version
        )  # pylint: disable=protected-access


def _set_api_attributes(span):
    if not span.is_recording():
        return

    if is_openai_v1():
        return

    base_url = openai.base_url if hasattr(openai, "base_url") else openai.api_base

    _set_span_attribute(span, SpanAttributes.LLM_OPENAI_API_BASE, base_url)
    _set_span_attribute(span, SpanAttributes.LLM_OPENAI_API_TYPE, openai.api_type)
    _set_span_attribute(span, SpanAttributes.LLM_OPENAI_API_VERSION, openai.api_version)

    return


def _set_functions_attributes(span, functions):
    if not functions:
        return

    for i, function in enumerate(functions):
        prefix = f"{SpanAttributes.LLM_REQUEST_FUNCTIONS}.{i}"
        _set_span_attribute(span, f"{prefix}.name", function.get("name"))
        _set_span_attribute(span, f"{prefix}.description", function.get("description"))
        _set_span_attribute(
            span, f"{prefix}.parameters", json.dumps(function.get("parameters"))
        )


def set_tools_attributes(span, tools):
    if not tools:
        return

    for i, tool in enumerate(tools):
        function = tool.get("function")
        if not function:
            continue

        prefix = f"{SpanAttributes.LLM_REQUEST_FUNCTIONS}.{i}"
        _set_span_attribute(span, f"{prefix}.name", function.get("name"))
        _set_span_attribute(span, f"{prefix}.description", function.get("description"))
        _set_span_attribute(
            span, f"{prefix}.parameters", json.dumps(function.get("parameters"))
        )


def _set_request_attributes(span, kwargs):
    if not span.is_recording():
        return

    _set_api_attributes(span)
    _set_span_attribute(span, SpanAttributes.LLM_SYSTEM, "OpenAI")
    _set_span_attribute(span, SpanAttributes.LLM_REQUEST_MODEL, kwargs.get("model"))
    _set_span_attribute(
        span, SpanAttributes.LLM_REQUEST_MAX_TOKENS, kwargs.get("max_tokens")
    )
    _set_span_attribute(
        span, SpanAttributes.LLM_REQUEST_TEMPERATURE, kwargs.get("temperature")
    )
    _set_span_attribute(span, SpanAttributes.LLM_REQUEST_TOP_P, kwargs.get("top_p"))
    _set_span_attribute(
        span, SpanAttributes.LLM_FREQUENCY_PENALTY, kwargs.get("frequency_penalty")
    )
    _set_span_attribute(
        span, SpanAttributes.LLM_PRESENCE_PENALTY, kwargs.get("presence_penalty")
    )
    _set_span_attribute(span, SpanAttributes.LLM_USER, kwargs.get("user"))
    _set_span_attribute(span, SpanAttributes.LLM_HEADERS, str(kwargs.get("headers")))
    # The new OpenAI SDK removed the `headers` and create new field called `extra_headers`
    if kwargs.get("extra_headers") is not None:
        _set_span_attribute(
            span, SpanAttributes.LLM_HEADERS, str(kwargs.get("extra_headers"))
        )
    _set_span_attribute(
        span, SpanAttributes.LLM_IS_STREAMING, kwargs.get("stream") or False
    )


@dont_throw
def _set_response_attributes(span, response):
    if not span.is_recording():
        logger.warning("Failed to set attributes, OpenTelemetry span is not recording")
        return

    _set_span_attribute(span, SpanAttributes.LLM_RESPONSE_MODEL, response.get("model"))

    _set_span_attribute(
        span,
        SpanAttributes.LLM_OPENAI_RESPONSE_SYSTEM_FINGERPRINT,
        response.get("system_fingerprint"),
    )

    usage = response.get("usage")
    if not usage:
        return

    if is_openai_v1() and not isinstance(usage, dict):
        usage = usage.__dict__

    _set_span_attribute(
        span, SpanAttributes.LLM_USAGE_TOTAL_TOKENS, usage.get("total_tokens")
    )
    _set_span_attribute(
        span,
        SpanAttributes.LLM_USAGE_COMPLETION_TOKENS,
        usage.get("completion_tokens"),
    )
    _set_span_attribute(
        span, SpanAttributes.LLM_USAGE_PROMPT_TOKENS, usage.get("prompt_tokens")
    )

    return


@dont_throw
def _set_span_stream_usage(span, prompt_tokens, completion_tokens):
    if not span.is_recording():
        return

    if isinstance(completion_tokens, int) and completion_tokens >= 0:
        _set_span_attribute(
            span, SpanAttributes.LLM_USAGE_COMPLETION_TOKENS, completion_tokens
        )

    if isinstance(prompt_tokens, int) and prompt_tokens >= 0:
        _set_span_attribute(span, SpanAttributes.LLM_USAGE_PROMPT_TOKENS, prompt_tokens)

    if (
        isinstance(prompt_tokens, int)
        and isinstance(completion_tokens, int)
        and completion_tokens + prompt_tokens >= 0
    ):
        _set_span_attribute(
            span,
            SpanAttributes.LLM_USAGE_TOTAL_TOKENS,
            completion_tokens + prompt_tokens,
        )


def _get_openai_base_url(instance):
    if hasattr(instance, "_client"):
        client = instance._client  # pylint: disable=protected-access
        if isinstance(client, (openai.AsyncOpenAI, openai.OpenAI)):
            return str(client.base_url)

    return ""


def is_streaming_response(response):
    if is_openai_v1():
        return isinstance(response, openai.Stream) or isinstance(
            response, openai.AsyncStream
        )

    return isinstance(response, types.GeneratorType) or isinstance(
        response, types.AsyncGeneratorType
    )


def model_as_dict(model):
    if version("pydantic") < "2.0.0":
        return model.dict()
    if hasattr(model, "model_dump"):
        return model.model_dump()
    elif hasattr(model, "parse"):  # Raw API response
        return model_as_dict(model.parse())
    else:
        return model


def get_token_count_from_string(string: str, model_name: str):
    if not should_record_stream_token_usage():
        return None

    import tiktoken

    if tiktoken_encodings.get(model_name) is None:
        try:
            encoding = tiktoken.encoding_for_model(model_name)
        except KeyError as e:
            # no such model_name in tiktoken
            logger.debug(
                f"Failed to get tiktoken encoding for model_name {model_name}, defaulting to cl100k_base (gpt-4). Error:\n{str(e)}"
            )
            encoding = tiktoken.encoding_for_model(DEFAULT_MODEL_FOR_ENCODING)
        except Exception as e:
            logger.exception(e)
            encoding = tiktoken.encoding_for_model(DEFAULT_MODEL_FOR_ENCODING)

        tiktoken_encodings[model_name] = encoding
    else:
        encoding = tiktoken_encodings.get(model_name)

    if not encoding:
        return None

    token_count = len(encoding.encode(string))
    return token_count


def _token_type(token_type: str):
    if token_type == "prompt_tokens":
        return "input"
    elif token_type == "completion_tokens":
        return "output"

    return None


def _metric_shared_attributes(
    response_model: str, operation: str, server_address: str, is_streaming: bool = False
):
    return {
        SpanAttributes.LLM_SYSTEM: "openai",
        SpanAttributes.LLM_RESPONSE_MODEL: response_model,
        "gen_ai.operation.name": operation,
        "server.address": server_address,
        "stream": is_streaming,
    }



from opentelemetry.sdk._logs import LoggingHandler


def monkeypatch_print(logging_handler: LoggingHandler):
    import builtins
    import inspect
    import logging
    import sys
    import traceback
    from collections.abc import Mapping

    print_logger = logging.getLogger("print")
    print_logger.setLevel(logging.INFO)
    print_logger.propagate = False

    # Attach print-style formatter (i.e., no LEVEL: prefix)
    print_logger.addHandler(logging_handler)

    def custom_print(
        *objects,
        sep=" ",
        end="\n",
        file=sys.stdout,
        flush=False,
        extra: Mapping[str, object] | None = None
    ):
        level = logging.ERROR if file is sys.stderr else logging.INFO
        message = sep.join(map(str, objects)) + end
        frame = inspect.currentframe()

        # Move up in the stack to find the first non-monkeypatch frame (the original caller of print)
        while frame and inspect.getsourcefile(frame) == __file__:
            frame = frame.f_back

        if frame:
            fn = frame.f_code.co_filename
            lno = frame.f_lineno
            func = frame.f_code.co_name
            # Capture stack trace from caller's frame
            sinfo = "".join(traceback.format_stack(frame))
        else:
            fn, lno, func, sinfo = "(unknown)", 0, "(unknown)", None

        record = print_logger.makeRecord(
            name=print_logger.name,
            level=level,
            fn=fn,
            lno=lno,
            msg=message,
            args=(),  # print API doesn't use this
            exc_info=None,  # print API doesn't use this
            func=func,
            extra=extra,
            sinfo=sinfo,
        )
        print_logger.handle(record)

        # Write to the file if needed
        if file is sys.stdout or file is sys.stderr:
            file.write(message)
            if flush:
                file.flush()

    builtins.print = custom_print



import logging
from typing import Optional, Sequence

import opentelemetry.attributes
from opentelemetry.attributes import _clean_attribute_value
from opentelemetry.util import types

logger = logging.getLogger(__name__)


_VALID_ATTR_VALUE_TYPES = (bool, str, bytes, int, float)


def patched_clean_attribute(
    key: str, value: types.AttributeValue, max_len: Optional[int]
) -> Optional[types.AttributeValue]:
    """Checks if attribute value is valid and cleans it if required.

    The function returns the cleaned value or None if the value is not valid.

    An attribute value is valid if it is either:
        - A primitive type: string, boolean, double precision floating
            point (IEEE 754-1985) or integer.
        - An array of primitive type values. The array MUST be homogeneous,
            i.e. it MUST NOT contain values of different types.

    An attribute needs cleansing if:
        - Its length is greater than the maximum allowed length.
        - It needs to be encoded/decoded e.g, bytes to strings.
    """

    if not (key and isinstance(key, str)):
        logger.warning("invalid key `%s`. must be non-empty string.", key)
        return None

    if isinstance(value, _VALID_ATTR_VALUE_TYPES):
        return _clean_attribute_value(value, max_len)

    if isinstance(value, Sequence):
        sequence_first_valid_type = None
        cleaned_seq = []

        for element in value:
            element = _clean_attribute_value(element, max_len)
            if element is None:
                cleaned_seq.append(element)
                continue

            element_type = type(element)
            # Reject attribute value if sequence contains a value with an incompatible type.
            if element_type not in _VALID_ATTR_VALUE_TYPES:
                logger.warning(
                    "Invalid type %s in attribute '%s' value sequence. Expected one of "
                    "%s or None",
                    element_type.__name__,
                    key,
                    [valid_type.__name__ for valid_type in _VALID_ATTR_VALUE_TYPES],
                )
                return None

            # The type of the sequence must be homogeneous. The first non-None
            # element determines the type of the sequence
            if sequence_first_valid_type is None:
                sequence_first_valid_type = element_type
            # use equality instead of isinstance as isinstance(True, int) evaluates to True
            elif element_type != sequence_first_valid_type:
                logger.warning(
                    "Attribute %r mixes types %s and %s in attribute value sequence",
                    key,
                    sequence_first_valid_type.__name__,
                    type(element).__name__,
                )
                return None

            cleaned_seq.append(element)

        # Freeze mutable sequences defensively
        return tuple(cleaned_seq)

    # Fall back to stringify (dict, error, etc.)
    # Rather than warning and returning None as before
    return _clean_attribute_value(str(value), max_len)


def monkeypatch_clean_attribute():
    opentelemetry.attributes._clean_attribute = patched_clean_attribute



import json
import logging
import time

from opentelemetry import context as context_api
from opentelemetry.context import _SUPPRESS_INSTRUMENTATION_KEY
from opentelemetry.metrics import Counter, Histogram
from opentelemetry.semconv.ai import LLMRequestTypeValues, SpanAttributes
from opentelemetry.trace import SpanKind, Tracer
from opentelemetry.trace.status import Status, StatusCode
from wrapt import ObjectProxy

from iudex.openai.shared import DEFAULT_MODEL_FOR_ENCODING

from ..shared import (
    OPENAI_LLM_USAGE_TOKEN_TYPES,
    _get_openai_base_url,
    _metric_shared_attributes,
    _set_client_attributes,
    _set_functions_attributes,
    _set_request_attributes,
    _set_response_attributes,
    _set_span_attribute,
    _set_span_stream_usage,
    _token_type,
    get_token_count_from_string,
    is_streaming_response,
    model_as_dict,
    set_tools_attributes,
    should_record_stream_token_usage,
    should_send_prompts,
)
from ..utils import (
    _with_chat_telemetry_wrapper,
    dont_throw,
    is_azure_openai,
    is_openai_v1,
)

SPAN_NAME = "openai.chat"
LLM_REQUEST_TYPE = LLMRequestTypeValues.CHAT

logger = logging.getLogger(__name__)


@_with_chat_telemetry_wrapper
def chat_wrapper(
    tracer: Tracer,
    token_counter: Counter,
    choice_counter: Counter,
    duration_histogram: Histogram,
    exception_counter: Counter,
    streaming_time_to_first_token: Histogram,
    streaming_time_to_generate: Histogram,
    wrapped,
    instance,
    args,
    kwargs,
):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    # span needs to be opened and closed manually because the response is a generator
    span = tracer.start_span(
        SPAN_NAME,
        kind=SpanKind.CLIENT,
        attributes={SpanAttributes.LLM_REQUEST_TYPE: LLM_REQUEST_TYPE.value},
    )

    _handle_request(span, kwargs, instance)

    try:
        start_time = time.time()
        response = wrapped(*args, **kwargs)
        end_time = time.time()
    except Exception as e:  # pylint: disable=broad-except
        end_time = time.time()
        duration = end_time - start_time if "start_time" in locals() else 0

        attributes = {
            "error.type": e.__class__.__name__,
        }

        if duration > 0 and duration_histogram:
            duration_histogram.record(duration, attributes=attributes)
        if exception_counter:
            exception_counter.add(1, attributes=attributes)

        raise e

    if is_streaming_response(response):
        # span will be closed after the generator is done
        if is_openai_v1():
            return ChatStream(
                span,
                response,
                instance,
                token_counter,
                choice_counter,
                duration_histogram,
                streaming_time_to_first_token,
                streaming_time_to_generate,
                start_time,
                kwargs,
            )
        else:
            return _build_from_streaming_response(
                span,
                response,
                instance,
                token_counter,
                choice_counter,
                duration_histogram,
                streaming_time_to_first_token,
                streaming_time_to_generate,
                start_time,
                kwargs,
            )

    duration = end_time - start_time

    _handle_response(
        response,
        span,
        instance,
        token_counter,
        choice_counter,
        duration_histogram,
        duration,
    )
    span.end()

    return response


@_with_chat_telemetry_wrapper
async def achat_wrapper(
    tracer: Tracer,
    token_counter: Counter,
    choice_counter: Counter,
    duration_histogram: Histogram,
    exception_counter: Counter,
    streaming_time_to_first_token: Histogram,
    streaming_time_to_generate: Histogram,
    wrapped,
    instance,
    args,
    kwargs,
):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    span = tracer.start_span(
        SPAN_NAME,
        kind=SpanKind.CLIENT,
        attributes={SpanAttributes.LLM_REQUEST_TYPE: LLM_REQUEST_TYPE.value},
    )
    _handle_request(span, kwargs, instance)

    try:
        start_time = time.time()
        response = await wrapped(*args, **kwargs)
        end_time = time.time()
    except Exception as e:  # pylint: disable=broad-except
        end_time = time.time()
        duration = end_time - start_time if "start_time" in locals() else 0

        attributes = {
            "error.type": e.__class__.__name__,
        }

        if duration > 0 and duration_histogram:
            duration_histogram.record(duration, attributes=attributes)
        if exception_counter:
            exception_counter.add(1, attributes=attributes)

        raise e

    if is_streaming_response(response):
        # span will be closed after the generator is done
        if is_openai_v1():
            return ChatStream(
                span,
                response,
                instance,
                token_counter,
                choice_counter,
                duration_histogram,
                streaming_time_to_first_token,
                streaming_time_to_generate,
                start_time,
                kwargs,
            )
        else:
            return _abuild_from_streaming_response(
                span,
                response,
                instance,
                token_counter,
                choice_counter,
                duration_histogram,
                streaming_time_to_first_token,
                streaming_time_to_generate,
                start_time,
                kwargs,
            )

    duration = end_time - start_time

    _handle_response(
        response,
        span,
        instance,
        token_counter,
        choice_counter,
        duration_histogram,
        duration,
    )
    span.end()

    return response


@dont_throw
def _handle_request(span, kwargs, instance):
    _set_request_attributes(span, kwargs)
    _set_client_attributes(span, instance)
    if should_send_prompts():
        _set_prompts(span, kwargs.get("messages"))
        if kwargs.get("functions"):
            _set_functions_attributes(span, kwargs.get("functions"))
        elif kwargs.get("tools"):
            set_tools_attributes(span, kwargs.get("tools"))


@dont_throw
def _handle_response(
    response,
    span,
    instance=None,
    token_counter=None,
    choice_counter=None,
    duration_histogram=None,
    duration=None,
):
    if is_openai_v1():
        response_dict = model_as_dict(response)
    else:
        response_dict = response

    # metrics record
    _set_chat_metrics(
        instance,
        token_counter,
        choice_counter,
        duration_histogram,
        response_dict,
        duration,
    )

    # span attributes
    _set_response_attributes(span, response_dict)

    if should_send_prompts():
        _set_completions(span, response_dict.get("choices"))

    return response


def _set_chat_metrics(
    instance, token_counter, choice_counter, duration_histogram, response_dict, duration
):
    shared_attributes = _metric_shared_attributes(
        response_model=response_dict.get("model") or None,
        operation="chat",
        server_address=_get_openai_base_url(instance),
        is_streaming=False,
    )

    # token metrics
    usage = response_dict.get("usage")  # type: dict
    if usage and token_counter:
        _set_token_counter_metrics(token_counter, usage, shared_attributes)

    # choices metrics
    choices = response_dict.get("choices")
    if choices and choice_counter:
        _set_choice_counter_metrics(choice_counter, choices, shared_attributes)

    # duration metrics
    if duration and isinstance(duration, (float, int)) and duration_histogram:
        duration_histogram.record(duration, attributes=shared_attributes)


def _set_choice_counter_metrics(choice_counter, choices, shared_attributes):
    choice_counter.add(len(choices), attributes=shared_attributes)
    for choice in choices:
        attributes_with_reason = {**shared_attributes}
        if choice.get("finish_reason"):
            attributes_with_reason[SpanAttributes.LLM_RESPONSE_FINISH_REASON] = (
                choice.get("finish_reason")
            )
        choice_counter.add(1, attributes=attributes_with_reason)


def _set_token_counter_metrics(token_counter, usage, shared_attributes):
    for name, val in usage.items():
        if name in OPENAI_LLM_USAGE_TOKEN_TYPES:
            attributes_with_token_type = {
                **shared_attributes,
                SpanAttributes.LLM_TOKEN_TYPE: _token_type(name),
            }
            token_counter.record(val, attributes=attributes_with_token_type)


def _set_prompts(span, messages):
    if not span.is_recording() or messages is None:
        return

    for i, msg in enumerate(messages):
        prefix = f"{SpanAttributes.LLM_PROMPTS}.{i}"
        if isinstance(msg.get("content"), str):
            content = msg.get("content")
        elif isinstance(msg.get("content"), list):
            content = json.dumps(msg.get("content"))

        _set_span_attribute(span, f"{prefix}.role", msg.get("role"))
        _set_span_attribute(span, f"{prefix}.content", content)


def _set_completions(span, choices):
    if choices is None:
        return

    for choice in choices:
        index = choice.get("index")
        prefix = f"{SpanAttributes.LLM_COMPLETIONS}.{index}"
        _set_span_attribute(
            span, f"{prefix}.finish_reason", choice.get("finish_reason")
        )

        if choice.get("finish_reason") == "content_filter":
            _set_span_attribute(span, f"{prefix}.role", "assistant")
            _set_span_attribute(span, f"{prefix}.content", "FILTERED")
            return

        message = choice.get("message")
        if not message:
            return

        _set_span_attribute(span, f"{prefix}.role", message.get("role"))
        _set_span_attribute(span, f"{prefix}.content", message.get("content"))

        function_call = message.get("function_call")
        if function_call:
            _set_span_attribute(
                span, f"{prefix}.function_call.name", function_call.get("name")
            )
            _set_span_attribute(
                span,
                f"{prefix}.function_call.arguments",
                function_call.get("arguments"),
            )

        tool_calls = message.get("tool_calls")
        if tool_calls:
            _set_span_attribute(
                span,
                f"{prefix}.function_call.name",
                tool_calls[0].get("function").get("name"),
            )
            _set_span_attribute(
                span,
                f"{prefix}.function_call.arguments",
                tool_calls[0].get("function").get("arguments"),
            )


def _set_streaming_token_metrics(
    request_kwargs, complete_response, span, token_counter, shared_attributes
):
    # use tiktoken calculate token usage
    if not should_record_stream_token_usage():
        return

    # kwargs={'model': 'gpt-3.5', 'messages': [{'role': 'user', 'content': '...'}], 'stream': True}
    prompt_usage = -1
    completion_usage = -1

    usage = complete_response.get("usage")
    choices = complete_response.get("choices")

    # parse streaming usage if present
    # https://platform.openai.com/docs/api-reference/chat/create#chat-create-stream_options
    if usage and not choices:
        prompt_usage = usage.get("prompt_tokens", -1)
        completion_usage = usage.get("completion_tokens", -1)

    # prompt_usage from tokenized content, if no usage
    if prompt_usage == -1 and request_kwargs and request_kwargs.get("messages"):
        prompt_content = ""
        # setting the default model_name as gpt-4. As this uses the embedding "cl100k_base" that
        # is used by most of the other model.
        model_name = (
            request_kwargs.get("model") or complete_response.get("model") or DEFAULT_MODEL_FOR_ENCODING
        )
        for msg in request_kwargs.get("messages"):
            if msg.get("content"):
                prompt_content += msg.get("content")
        if model_name:
            prompt_usage = get_token_count_from_string(prompt_content, model_name)

    # completion_usage from tokenized content, if no usage
    if completion_usage == -1 and choices:
        completion_content = ""
        # setting the default model_name as gpt-4. As this uses the embedding "cl100k_base" that
        # is used by most of the other model.
        model_name = complete_response.get("model") or DEFAULT_MODEL_FOR_ENCODING

        for choice in choices:
            if choice.get("message") and choice.get("message").get("content"):
                completion_content += choice["message"]["content"]

        if model_name:
            completion_usage = get_token_count_from_string(
                completion_content, model_name
            )

    # span record
    _set_span_stream_usage(span, prompt_usage, completion_usage)

    # metrics record
    if token_counter:
        if isinstance(prompt_usage, int) and prompt_usage >= 0:
            attributes_with_token_type = {
                **shared_attributes,
                SpanAttributes.LLM_TOKEN_TYPE: "input",
            }
            token_counter.record(prompt_usage, attributes=attributes_with_token_type)

        if isinstance(completion_usage, int)  and completion_usage >= 0:
            attributes_with_token_type = {
                **shared_attributes,
                SpanAttributes.LLM_TOKEN_TYPE: "output",
            }
            token_counter.record(
                completion_usage, attributes=attributes_with_token_type
            )


class ChatStream(ObjectProxy):
    _span = None
    _instance = None
    _token_counter = None
    _choice_counter = None
    _duration_histogram = None
    _streaming_time_to_first_token = None
    _streaming_time_to_generate = None
    _start_time = None
    _request_kwargs = None

    def __init__(
        self,
        span,
        response,
        instance=None,
        token_counter=None,
        choice_counter=None,
        duration_histogram=None,
        streaming_time_to_first_token=None,
        streaming_time_to_generate=None,
        start_time=None,
        request_kwargs=None,
    ):
        super().__init__(response)

        self._span = span
        self._instance = instance
        self._token_counter = token_counter
        self._choice_counter = choice_counter
        self._duration_histogram = duration_histogram
        self._streaming_time_to_first_token = streaming_time_to_first_token
        self._streaming_time_to_generate = streaming_time_to_generate
        self._start_time = start_time
        self._request_kwargs = request_kwargs

        self._first_token = True
        # will be updated when first token is received
        self._time_of_first_token = self._start_time
        self._complete_response = {"choices": [], "model": ""}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__wrapped__.__exit__(exc_type, exc_val, exc_tb)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.__wrapped__.__aexit__(exc_type, exc_val, exc_tb)

    def __iter__(self):
        return self

    def __aiter__(self):
        return self

    def __next__(self):
        try:
            chunk = self.__wrapped__.__next__()
        except Exception as e:
            if isinstance(e, StopIteration):
                self._close_span()
            raise e
        else:
            self._process_item(chunk)
            return chunk

    async def __anext__(self):
        try:
            chunk = await self.__wrapped__.__anext__()
        except Exception as e:
            if isinstance(e, StopAsyncIteration):
                self._close_span()
            raise e
        else:
            self._process_item(chunk)
            return chunk

    def _process_item(self, item):
        self._span.add_event(name=f"{SpanAttributes.LLM_CONTENT_COMPLETION_CHUNK}")

        if self._first_token and self._streaming_time_to_first_token:
            self._time_of_first_token = time.time()
            self._streaming_time_to_first_token.record(
                self._time_of_first_token - self._start_time,
                attributes=self._shared_attributes(),
            )
            self._first_token = False

        if is_openai_v1():
            item = model_as_dict(item)

        self._complete_response["model"] = item.get("model")

        for choice in item.get("choices"):
            index = choice.get("index")
            # prevent weird index values
            if index is None or index > 1000:
                logger.warning(f"Skipping invalid choice with index: {index}")
                continue
            while len(self._complete_response["choices"]) <= index:
                self._complete_response["choices"].append(
                    {"index": index, "message": {"content": "", "role": ""}}
                )
            try:
                complete_choice = self._complete_response["choices"][index]
            except IndexError as e:
                logger.exception(f"Error processing choice with index {index}: {e}")
                continue
            if choice.get("finish_reason"):
                complete_choice["finish_reason"] = choice.get("finish_reason")

            delta = choice.get("delta")

            if delta and delta.get("content"):
                complete_choice["message"]["content"] += delta.get("content")
            if delta and delta.get("role"):
                complete_choice["message"]["role"] = delta.get("role")
            if delta and delta.get("tool_calls"):
                tool_calls = delta.get("tool_calls")
                if not isinstance(tool_calls, list) or len(tool_calls) == 0:
                    continue

                if not complete_choice["message"].get("tool_calls"):
                    complete_choice["message"]["tool_calls"] = [
                        {"function": {"name": "", "arguments": ""}}
                    ]

                tool_call = tool_calls[0]
                function = complete_choice["message"]["tool_calls"][0]["function"]

                if tool_call.get("function") and tool_call["function"].get("name"):
                    function["name"] += tool_call["function"]["name"]
                if tool_call.get("function") and tool_call["function"].get("arguments"):
                    function["arguments"] += tool_call["function"]["arguments"]

    def _shared_attributes(self):
        return _metric_shared_attributes(
            response_model=self._complete_response.get("model")
            or self._request_kwargs.get("model")
            or None,
            operation="chat",
            server_address=_get_openai_base_url(self._instance),
            is_streaming=True,
        )

    @dont_throw
    def _close_span(self):
        _set_streaming_token_metrics(
            self._request_kwargs,
            self._complete_response,
            self._span,
            self._token_counter,
            self._shared_attributes(),
        )

        # choice metrics
        if self._choice_counter and self._complete_response.get("choices"):
            _set_choice_counter_metrics(
                self._choice_counter,
                self._complete_response.get("choices"),
                self._shared_attributes(),
            )

        # duration metrics
        if self._start_time and isinstance(self._start_time, (float, int)):
            duration = time.time() - self._start_time
        else:
            duration = None
        if duration and isinstance(duration, (float, int)) and self._duration_histogram:
            self._duration_histogram.record(
                duration, attributes=self._shared_attributes()
            )
        if self._streaming_time_to_generate and self._time_of_first_token:
            self._streaming_time_to_generate.record(
                time.time() - self._time_of_first_token,
                attributes=self._shared_attributes(),
            )

        _set_response_attributes(self._span, self._complete_response)

        if should_send_prompts():
            _set_completions(self._span, self._complete_response.get("choices"))

        self._span.set_status(Status(StatusCode.OK))
        self._span.end()


# Backward compatibility with OpenAI v0


@dont_throw
def _build_from_streaming_response(
    span,
    response,
    instance=None,
    token_counter=None,
    choice_counter=None,
    duration_histogram=None,
    streaming_time_to_first_token=None,
    streaming_time_to_generate=None,
    start_time=None,
    request_kwargs=None,
):
    complete_response = {"choices": [], "model": ""}

    first_token = True
    time_of_first_token = start_time  # will be updated when first token is received

    for item in response:
        span.add_event(name=f"{SpanAttributes.LLM_CONTENT_COMPLETION_CHUNK}")

        item_to_yield = item

        if first_token and streaming_time_to_first_token:
            time_of_first_token = time.time()
            streaming_time_to_first_token.record(time_of_first_token - start_time)
            first_token = False

        _accumulate_stream_items(item, complete_response)

        yield item_to_yield

    shared_attributes = {
        SpanAttributes.LLM_RESPONSE_MODEL: complete_response.get("model") or None,
        "server.address": _get_openai_base_url(instance),
        "stream": True,
    }

    _set_streaming_token_metrics(
        request_kwargs, complete_response, span, token_counter, shared_attributes
    )

    # choice metrics
    if choice_counter and complete_response.get("choices"):
        _set_choice_counter_metrics(
            choice_counter, complete_response.get("choices"), shared_attributes
        )

    # duration metrics
    if start_time and isinstance(start_time, (float, int)):
        duration = time.time() - start_time
    else:
        duration = None
    if duration and isinstance(duration, (float, int)) and duration_histogram:
        duration_histogram.record(duration, attributes=shared_attributes)
    if streaming_time_to_generate and time_of_first_token:
        streaming_time_to_generate.record(time.time() - time_of_first_token)

    _set_response_attributes(span, complete_response)

    if should_send_prompts():
        _set_completions(span, complete_response.get("choices"))

    span.set_status(Status(StatusCode.OK))
    span.end()


@dont_throw
async def _abuild_from_streaming_response(
    span,
    response,
    instance=None,
    token_counter=None,
    choice_counter=None,
    duration_histogram=None,
    streaming_time_to_first_token=None,
    streaming_time_to_generate=None,
    start_time=None,
    request_kwargs=None,
):
    complete_response = {"choices": [], "model": ""}

    first_token = True
    time_of_first_token = start_time  # will be updated when first token is received

    async for item in response:
        span.add_event(name=f"{SpanAttributes.LLM_CONTENT_COMPLETION_CHUNK}")

        item_to_yield = item

        if first_token and streaming_time_to_first_token:
            time_of_first_token = time.time()
            streaming_time_to_first_token.record(time_of_first_token - start_time)
            first_token = False

        _accumulate_stream_items(item, complete_response)

        yield item_to_yield

    shared_attributes = {
        SpanAttributes.LLM_RESPONSE_MODEL: complete_response.get("model") or None,
        "server.address": _get_openai_base_url(instance),
        "stream": True,
    }

    if not is_azure_openai(instance):
        _set_streaming_token_metrics(
            request_kwargs, complete_response, span, token_counter, shared_attributes
        )

    # choice metrics
    if choice_counter and complete_response.get("choices"):
        _set_choice_counter_metrics(
            choice_counter, complete_response.get("choices"), shared_attributes
        )

    # duration metrics
    if start_time and isinstance(start_time, (float, int)):
        duration = time.time() - start_time
    else:
        duration = None
    if duration and isinstance(duration, (float, int)) and duration_histogram:
        duration_histogram.record(duration, attributes=shared_attributes)
    if streaming_time_to_generate and time_of_first_token:
        streaming_time_to_generate.record(time.time() - time_of_first_token)

    _set_response_attributes(span, complete_response)

    if should_send_prompts():
        _set_completions(span, complete_response.get("choices"))

    span.set_status(Status(StatusCode.OK))
    span.end()


def _accumulate_stream_items(item, complete_response):
    if is_openai_v1():
        item = model_as_dict(item)

    complete_response["model"] = item.get("model")

    for choice in item.get("choices"):
        index = choice.get("index")
        if len(complete_response.get("choices")) <= index:
            complete_response["choices"].append(
                {"index": index, "message": {"content": "", "role": ""}}
            )
        complete_choice = complete_response.get("choices")[index]
        if choice.get("finish_reason"):
            complete_choice["finish_reason"] = choice.get("finish_reason")

        delta = choice.get("delta")

        if delta and delta.get("content"):
            complete_choice["message"]["content"] += delta.get("content")
        if delta and delta.get("role"):
            complete_choice["message"]["role"] = delta.get("role")



from typing import Collection

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.metrics import get_meter
from opentelemetry.semconv.ai import Meters
from opentelemetry.trace import get_tracer
from wrapt import wrap_function_wrapper

from ..shared.chat_wrappers import achat_wrapper, chat_wrapper
from ..shared.completion_wrappers import acompletion_wrapper, completion_wrapper
from ..shared.embeddings_wrappers import aembeddings_wrapper, embeddings_wrapper
from ..shared.image_gen_wrappers import image_gen_metrics_wrapper
from ..utils import is_metrics_enabled
from ..version import __version__
from .assistant_wrappers import (
    assistants_create_wrapper,
    messages_list_wrapper,
    runs_create_and_stream_wrapper,
    runs_create_wrapper,
    runs_retrieve_wrapper,
)

_instruments = ("openai >= 1.0.0",)


class OpenAIV1Instrumentor(BaseInstrumentor):
    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, __version__, tracer_provider)

        # meter and counters are inited here
        meter_provider = kwargs.get("meter_provider")
        meter = get_meter(__name__, __version__, meter_provider)

        if is_metrics_enabled():
            tokens_histogram = meter.create_histogram(
                name=Meters.LLM_TOKEN_USAGE,
                unit="token",
                description="Measures number of input and output tokens used",
            )

            chat_choice_counter = meter.create_counter(
                name=Meters.LLM_GENERATION_CHOICES,
                unit="choice",
                description="Number of choices returned by chat completions call",
            )

            duration_histogram = meter.create_histogram(
                name=Meters.LLM_OPERATION_DURATION,
                unit="s",
                description="GenAI operation duration",
            )

            chat_exception_counter = meter.create_counter(
                name=Meters.LLM_COMPLETIONS_EXCEPTIONS,
                unit="time",
                description="Number of exceptions occurred during chat completions",
            )

            streaming_time_to_first_token = meter.create_histogram(
                name=Meters.LLM_STREAMING_TIME_TO_FIRST_TOKEN,
                unit="s",
                description="Time to first token in streaming chat completions",
            )
            streaming_time_to_generate = meter.create_histogram(
                name=Meters.LLM_STREAMING_TIME_TO_GENERATE,
                unit="s",
                description="Time between first token and completion in streaming chat completions",
            )
        else:
            (
                tokens_histogram,
                chat_choice_counter,
                duration_histogram,
                chat_exception_counter,
                streaming_time_to_first_token,
                streaming_time_to_generate,
            ) = (None, None, None, None, None, None)

        wrap_function_wrapper(
            "openai.resources.chat.completions",
            "Completions.create",
            chat_wrapper(
                tracer,
                tokens_histogram,
                chat_choice_counter,
                duration_histogram,
                chat_exception_counter,
                streaming_time_to_first_token,
                streaming_time_to_generate,
            ),
        )

        wrap_function_wrapper(
            "openai.resources.completions",
            "Completions.create",
            completion_wrapper(tracer),
        )

        if is_metrics_enabled():
            embeddings_vector_size_counter = meter.create_counter(
                name=Meters.LLM_EMBEDDINGS_VECTOR_SIZE,
                unit="element",
                description="he size of returned vector",
            )
            embeddings_exception_counter = meter.create_counter(
                name=Meters.LLM_EMBEDDINGS_EXCEPTIONS,
                unit="time",
                description="Number of exceptions occurred during embeddings operation",
            )
        else:
            (
                tokens_histogram,
                embeddings_vector_size_counter,
                embeddings_exception_counter,
            ) = (None, None, None)

        wrap_function_wrapper(
            "openai.resources.embeddings",
            "Embeddings.create",
            embeddings_wrapper(
                tracer,
                tokens_histogram,
                embeddings_vector_size_counter,
                duration_histogram,
                embeddings_exception_counter,
            ),
        )

        wrap_function_wrapper(
            "openai.resources.chat.completions",
            "AsyncCompletions.create",
            achat_wrapper(
                tracer,
                tokens_histogram,
                chat_choice_counter,
                duration_histogram,
                chat_exception_counter,
                streaming_time_to_first_token,
                streaming_time_to_generate,
            ),
        )
        wrap_function_wrapper(
            "openai.resources.completions",
            "AsyncCompletions.create",
            acompletion_wrapper(tracer),
        )
        wrap_function_wrapper(
            "openai.resources.embeddings",
            "AsyncEmbeddings.create",
            aembeddings_wrapper(
                tracer,
                tokens_histogram,
                embeddings_vector_size_counter,
                duration_histogram,
                embeddings_exception_counter,
            ),
        )

        if is_metrics_enabled():
            image_gen_exception_counter = meter.create_counter(
                name=Meters.LLM_IMAGE_GENERATIONS_EXCEPTIONS,
                unit="time",
                description="Number of exceptions occurred during image generations operation",
            )
        else:
            image_gen_exception_counter = None

        wrap_function_wrapper(
            "openai.resources.images",
            "Images.generate",
            image_gen_metrics_wrapper(duration_histogram, image_gen_exception_counter),
        )

        # Beta APIs may not be available consistently in all versions
        try:
            wrap_function_wrapper(
                "openai.resources.beta.assistants",
                "Assistants.create",
                assistants_create_wrapper(tracer),
            )
            wrap_function_wrapper(
                "openai.resources.beta.threads.runs",
                "Runs.create",
                runs_create_wrapper(tracer),
            )
            wrap_function_wrapper(
                "openai.resources.beta.threads.runs",
                "Runs.retrieve",
                runs_retrieve_wrapper(tracer),
            )
            wrap_function_wrapper(
                "openai.resources.beta.threads.runs",
                "Runs.create_and_stream",
                runs_create_and_stream_wrapper(tracer),
            )
            wrap_function_wrapper(
                "openai.resources.beta.threads.messages",
                "Messages.list",
                messages_list_wrapper(tracer),
            )
        except AttributeError:
            pass

    def _uninstrument(self, **kwargs):
        pass



import logging
from collections.abc import Sequence
from typing import Any, Mapping

import opentelemetry.exporter.otlp.proto.common._internal
from opentelemetry.proto.common.v1.common_pb2 import AnyValue as PB2AnyValue
from opentelemetry.proto.common.v1.common_pb2 import (
    ArrayValue as PB2ArrayValue,
)
from opentelemetry.proto.common.v1.common_pb2 import KeyValue as PB2KeyValue
from opentelemetry.proto.common.v1.common_pb2 import (
    KeyValueList as PB2KeyValueList,
)


def patched_encode_value(value: Any):
    if isinstance(value, bool):
        return PB2AnyValue(bool_value=value)
    if isinstance(value, str):
        return PB2AnyValue(string_value=value)
    if isinstance(value, int):
        return PB2AnyValue(int_value=value)
    if isinstance(value, float):
        return PB2AnyValue(double_value=value)
    if isinstance(value, Sequence):
        return PB2AnyValue(
            array_value=PB2ArrayValue(
                values=[patched_encode_value(v) for v in value]
            )
        )
    elif isinstance(value, Mapping):
        return PB2AnyValue(
            kvlist_value=PB2KeyValueList(
                values=[
                    patched_encode_key_value(str(k), v) for k, v in value.items()
                ]
            )
        )
    try:
        return PB2AnyValue(string_value=str(value))
    except Exception as e:
        logging.exception(e)


def patched_encode_key_value(key: str, value: Any) -> PB2KeyValue:
    return PB2KeyValue(key=key, value=patched_encode_value(value))

def monkeypatch_encode_value():
    opentelemetry.exporter.otlp.proto.common._internal._encode_value = (
        patched_encode_value
    )
    opentelemetry.exporter.otlp.proto.common._internal._encode_key_value = (
        patched_encode_key_value
    )



import logging

from opentelemetry import context as context_api
from opentelemetry.context import _SUPPRESS_INSTRUMENTATION_KEY
from opentelemetry.semconv.ai import LLMRequestTypeValues, SpanAttributes
from opentelemetry.trace import SpanKind
from opentelemetry.trace.status import Status, StatusCode

from ..shared import (
    _set_client_attributes,
    _set_functions_attributes,
    _set_request_attributes,
    _set_response_attributes,
    _set_span_attribute,
    _set_span_stream_usage,
    get_token_count_from_string,
    is_streaming_response,
    model_as_dict,
    should_record_stream_token_usage,
    should_send_prompts,
)
from ..utils import _with_tracer_wrapper, dont_throw, is_openai_v1

SPAN_NAME = "openai.completion"
LLM_REQUEST_TYPE = LLMRequestTypeValues.COMPLETION

logger = logging.getLogger(__name__)


@_with_tracer_wrapper
def completion_wrapper(tracer, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    # span needs to be opened and closed manually because the response is a generator
    span = tracer.start_span(
        SPAN_NAME,
        kind=SpanKind.CLIENT,
        attributes={SpanAttributes.LLM_REQUEST_TYPE: LLM_REQUEST_TYPE.value},
    )

    _handle_request(span, kwargs, instance)
    response = wrapped(*args, **kwargs)

    if is_streaming_response(response):
        # span will be closed after the generator is done
        return _build_from_streaming_response(span, kwargs, response)
    else:
        _handle_response(response, span)

    span.end()
    return response


@_with_tracer_wrapper
async def acompletion_wrapper(tracer, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    span = tracer.start_span(
        name=SPAN_NAME,
        kind=SpanKind.CLIENT,
        attributes={SpanAttributes.LLM_REQUEST_TYPE: LLM_REQUEST_TYPE.value},
    )

    _handle_request(span, kwargs, instance)
    response = await wrapped(*args, **kwargs)

    if is_streaming_response(response):
        # span will be closed after the generator is done
        return _abuild_from_streaming_response(span, kwargs, response)
    else:
        _handle_response(response, span)

    span.end()
    return response


@dont_throw
def _handle_request(span, kwargs, instance):
    _set_request_attributes(span, kwargs)
    if should_send_prompts():
        _set_prompts(span, kwargs.get("prompt"))
        _set_functions_attributes(span, kwargs.get("functions"))
    _set_client_attributes(span, instance)


@dont_throw
def _handle_response(response, span):
    if is_openai_v1():
        response_dict = model_as_dict(response)
    else:
        response_dict = response

    _set_response_attributes(span, response_dict)

    if should_send_prompts():
        _set_completions(span, response_dict.get("choices"))


def _set_prompts(span, prompt):
    if not span.is_recording() or not prompt:
        return

    _set_span_attribute(
        span,
        f"{SpanAttributes.LLM_PROMPTS}.0.user",
        prompt[0] if isinstance(prompt, list) else prompt,
    )


@dont_throw
def _set_completions(span, choices):
    if not span.is_recording() or not choices:
        return

    for choice in choices:
        index = choice.get("index")
        prefix = f"{SpanAttributes.LLM_COMPLETIONS}.{index}"
        _set_span_attribute(
            span, f"{prefix}.finish_reason", choice.get("finish_reason")
        )
        _set_span_attribute(span, f"{prefix}.content", choice.get("text"))


@dont_throw
def _build_from_streaming_response(span, request_kwargs, response):
    complete_response = {"choices": [], "model": ""}
    for item in response:
        yield item
        _accumulate_streaming_response(complete_response, item)

    _set_response_attributes(span, complete_response)

    _set_token_usage(span, request_kwargs, complete_response)

    if should_send_prompts():
        _set_completions(span, complete_response.get("choices"))

    span.set_status(Status(StatusCode.OK))
    span.end()


@dont_throw
async def _abuild_from_streaming_response(span, request_kwargs, response):
    complete_response = {"choices": [], "model": ""}
    async for item in response:
        yield item
        _accumulate_streaming_response(complete_response, item)

    _set_response_attributes(span, complete_response)

    _set_token_usage(span, request_kwargs, complete_response)

    if should_send_prompts():
        _set_completions(span, complete_response.get("choices"))

    span.set_status(Status(StatusCode.OK))
    span.end()


@dont_throw
def _set_token_usage(span, request_kwargs, complete_response):
    # use tiktoken calculate token usage
    if should_record_stream_token_usage():
        prompt_usage = -1
        completion_usage = -1

        # prompt_usage
        if request_kwargs and request_kwargs.get("prompt"):
            prompt_content = request_kwargs.get("prompt")
            model_name = request_kwargs.get("model") or None

            if model_name:
                prompt_usage = get_token_count_from_string(prompt_content, model_name)

        # completion_usage
        if complete_response.get("choices"):
            completion_content = ""
            model_name = complete_response.get("model") or None

            for choice in complete_response.get("choices"):
                if choice.get("text"):
                    completion_content += choice.get("text")

            if model_name:
                completion_usage = get_token_count_from_string(
                    completion_content, model_name
                )

        # span record
        _set_span_stream_usage(span, prompt_usage, completion_usage)


@dont_throw
def _accumulate_streaming_response(complete_response, item):
    if is_openai_v1():
        item = model_as_dict(item)

    complete_response["model"] = item.get("model")

    for choice in item.get("choices"):
        index = choice.get("index")
        if len(complete_response.get("choices")) <= index:
            complete_response["choices"].append({"index": index, "text": ""})
        complete_choice = complete_response.get("choices")[index]
        if choice.get("finish_reason"):
            complete_choice["finish_reason"] = choice.get("finish_reason")

        if choice.get("text"):
            complete_choice["text"] += choice.get("text")

    return complete_response



from typing import Collection

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor

from .shared.config import Config
from .utils import is_openai_v1
from .v0 import OpenAIV0Instrumentor
from .v1 import OpenAIV1Instrumentor

_instruments = ("openai >= 0.27.0",)


class OpenAIInstrumentor(BaseInstrumentor):
    """An instrumentor for OpenAI's client library."""

    def __init__(
        self,
        enrich_assistant: bool = False,
        enrich_token_usage: bool = False,
        exception_logger=None,
    ):
        super().__init__()
        Config.enrich_assistant = enrich_assistant
        Config.enrich_token_usage = enrich_token_usage
        Config.exception_logger = exception_logger

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        if is_openai_v1():
            OpenAIV1Instrumentor().instrument(**kwargs)
        else:
            OpenAIV0Instrumentor().instrument(**kwargs)

    def _uninstrument(self, **kwargs):
        if is_openai_v1():
            OpenAIV1Instrumentor().uninstrument(**kwargs)
        else:
            OpenAIV0Instrumentor().uninstrument(**kwargs)



import logging
from typing import Dict, Literal, Optional

from httpx import Headers, QueryParams
from opentelemetry.util.types import AttributeValue

logger = logging.getLogger(__name__)

SqlCommand = Literal["SELECT", "INSERT", "UPDATE", "UPSERT", "DELETE"]

PARAM_TO_OPERATOR = {
    "eq": "eq",
    "neq": "neq",
    "gt": "gt",
    "gte": "gte",
    "lt": "lt",
    "lte": "lte",
    "like": "like",
    "ilike": "ilike",
    "is": "is_",
    "in": "in_",
    "cs": "contains",
    "cd": "contained_by",
    "match": "match",
    "not": "not_",
    # below are not used by supabase
    "imatch": "imatch",
    "isdistinct": "isdistinct",
    "fts": "fts",
    "plfts": "plfts",
    "phfts": "phfts",
    "wfts": "wfts",
    "ov": "ov",
    "sl": "sl",
    "sr": "sr",
    "nxr": "nxr",
    "nxl": "nxl",
    "adj": "adj",
    "or": "or",
    "and": "and",
    "all": "all",
    "any": "any",
}


def parse_table_name(path: str) -> Optional[str]:
    """Naively parse table name from Supabase path.

    Args:
        path: Request path, e.g. /my_table
    """
    try:
        return path.split("/")[1]
    except Exception as e:
        logger.exception("Failed to extract table name from path %s: %s", path, e)


def parse_sql_command(
    http_method: str, headers_attributes: Dict[str, AttributeValue]
) -> Optional[SqlCommand]:
    """Parse SQL command from Supabase request.

    Args:
        http_method: HTTP method, e.g. GET, POST, PUT, PATCH, DELETE
        headers_attributes: Parsed headers as attributes dict
    """
    if http_method == "GET" or http_method == "HEAD":
        return "SELECT"
    elif http_method == "POST":
        if (
            headers_attributes.get("http.header.prefer.resolution")
            == "merge-duplicates"
        ):
            return "UPSERT"
        return "INSERT"
    elif http_method == "PUT":
        return "UPSERT"
    elif http_method == "PATCH":
        return "UPDATE"
    elif http_method == "DELETE":
        return "DELETE"

    logger.exception(
        "Failed to parse SQL command from http method %s, headers %s",
        http_method,
        headers_attributes,
    )


def parse_prefer_header_to_attributes(headers: Headers) -> Dict[str, AttributeValue]:
    """Parse headers into OTel compliant attributes dict.

    Uses HTTP semconv.
    See https://opentelemetry.io/docs/specs/semconv/attributes-registry/http/#http-attributes.

    Also parses nested Prefer headers, e.g. Prefer: resolution=merge-duplicates
    See https://www.rfc-editor.org/rfc/rfc7240.html.

    Args:
        headers: HTTP headers
    """
    attributes = {}
    for key, value in headers.items():
        if key.lower() == "authorization":
            continue
        if key.lower() == "prefer":
            for prefer in value.split(","):
                prefer_key, prefer_value = prefer.split("=")
                attributes[f"http.header.prefer.{prefer_key}"] = prefer_value
            continue
        attributes[f"http.header.{key}"] = value
    return attributes


def parse_sql_statement(params: QueryParams) -> str:
    """Parse pseudo SQL statement from Supabase request params.

    See https://postgrest.org/en/v12/references/api/tables_views.html#operators

    We naively return the params in a syntax more similar to Supabase ORM than SQL.

    Ideally, we would start a span when the query begins building, monkey patch every
    subsequent query builder method, and collect the args as they are passed in.

    Args:
        params: Request params
    """
    statements = []
    for key, value in params.items():
        op = value.split(".")[0]
        if op in PARAM_TO_OPERATOR:
            statements.append(f".{PARAM_TO_OPERATOR[op]}({key}, {value[len(op) + 1:]})")
            continue
        statements.append(f".{key}({value})")
    return "".join(statements)



from openai import AssistantEventHandler
from opentelemetry.semconv.ai import SpanAttributes
from typing_extensions import override

from ..shared import _set_span_attribute


class EventHandleWrapper(AssistantEventHandler):

    _current_text_index = 0

    def __init__(self, original_handler, span):
        super().__init__()
        self._original_handler = original_handler
        self._span = span

    @override
    def on_end(self):
        self._original_handler.on_end()
        self._span.end()

    @override
    def on_event(self, event):
        self._original_handler.on_event(event)

    @override
    def on_run_step_created(self, run_step):
        self._original_handler.on_run_step_created(run_step)

    @override
    def on_run_step_delta(self, delta, snapshot):
        self._original_handler.on_run_step_delta(delta, snapshot)

    @override
    def on_run_step_done(self, run_step):
        self._original_handler.on_run_step_done(run_step)

    @override
    def on_tool_call_created(self, tool_call):
        self._original_handler.on_tool_call_created(tool_call)

    @override
    def on_tool_call_delta(self, delta, snapshot):
        self._original_handler.on_tool_call_delta(delta, snapshot)

    @override
    def on_tool_call_done(self, tool_call):
        self._original_handler.on_tool_call_done(tool_call)

    @override
    def on_exception(self, exception: Exception):
        self._original_handler.on_exception(exception)

    @override
    def on_timeout(self):
        self._original_handler.on_timeout()

    @override
    def on_message_created(self, message):
        self._original_handler.on_message_created(message)

    @override
    def on_message_delta(self, delta, snapshot):
        self._original_handler.on_message_delta(delta, snapshot)

    @override
    def on_message_done(self, message):
        self._original_handler.on_message_done(message)

    @override
    def on_text_created(self, text):
        self._original_handler.on_text_created(text)

    @override
    def on_text_delta(self, delta, snapshot):
        self._original_handler.on_text_delta(delta, snapshot)

    @override
    def on_text_done(self, text):
        self._original_handler.on_text_done(text)
        _set_span_attribute(
            self._span,
            f"{SpanAttributes.LLM_COMPLETIONS}.{self._current_text_index}.role",
            "assistant",
        )
        _set_span_attribute(
            self._span,
            f"{SpanAttributes.LLM_COMPLETIONS}.{self._current_text_index}.content",
            text.value,
        )

        self._current_text_index += 1

    @override
    def on_image_file_done(self, image_file):
        self._original_handler.on_image_file_done(image_file)



import importlib
import importlib.util
import importlib.metadata
from packaging.requirements import Requirement
import logging

logger = logging.getLogger(__name__)


def maybe_instrument_lib(module_path: str, instrumentor_class_name: str, **kwargs):
    try:
        # get instrumentor and its requirements
        package = "iudex" if module_path[0] == "." else None
        module = importlib.import_module(module_path, package)

        construct_instrumentor = getattr(module, instrumentor_class_name)
        instrumentor = construct_instrumentor(**kwargs)
        req_strs = instrumentor.instrumentation_dependencies()
        reqs = [Requirement(r) for r in req_strs]

        # check package requirements are satisfied
        for req in reqs:
            package_name = req.name

            if importlib.util.find_spec(package_name) is None:
                logger.debug(
                    f"Skipping {instrumentor_class_name} instrumentation as requirement {req.name} is not installed."
                )
                return

            package_version = importlib.metadata.version(package_name)
            if package_version not in req.specifier:
                raise ValueError(
                    f"Version {package_version} of {package_name} does not satisfy the requirement {req}."
                )

        if instrumentor.is_instrumented_by_opentelemetry:
            logger.debug(
                f"Skipping {instrumentor_class_name} instrumentation as it was already instrumented."
            )
            return

        # instrument the library
        instrumentor.instrument()

    # instrumentor tried and failed to import a requirement, so swallow error and skip
    except ModuleNotFoundError as e:
        logger.debug(f"Skipping {instrumentor_class_name} instrumentation: {e}")

    # edge case when AwsLambdaInstrumentor tries to rsplit on non-existent handler name
    except AttributeError as e:
        if str(e) == "'NoneType' object has no attribute 'rsplit'":
            logger.debug(f"Skipping {instrumentor_class_name} instrumentation: {e}")
            return
        logger.exception(f"Failed to instrument with {instrumentor_class_name}: {e}")

    except Exception as e:
        logger.exception(f"Failed to instrument with {instrumentor_class_name}: {e}")



# ruff: noqa: E402
# ^ because monkeypatches must import and run before other imports
from .monkeypatches.encode_value import monkeypatch_encode_value
from .monkeypatches.clean_attribute import monkeypatch_clean_attribute

monkeypatch_encode_value()
monkeypatch_clean_attribute()

from .config import IudexConfig, configure_logger
from .instrumentation import instrument
from .fastapi import instrument_fastapi
from .trace import trace, trace_lambda, start_trace, end_trace

__all__ = [
  "IudexConfig",
  "configure_logger",
  "instrument",
  "instrument_fastapi",
  "trace",
  "trace_lambda",
  "start_trace",
  "end_trace",
]



def monkeypatch_LogRecord_getMessage():
    """Monkey patches LogRecord.getMessages.

    Changes getMessages to also handle *args as a list of strings instead of only as
    a list of string format arguments.
    """
    from logging import LogRecord

    _getMessage = LogRecord.getMessage

    def getMessage(self: LogRecord):
        try:
            return _getMessage(self)
        except Exception:
            msg = str(self.msg)
            if self.args:
                msg = msg + " " + " ".join(self.args)
            return msg

    LogRecord.getMessage = getMessage



import logging
import time

from openai._legacy_response import LegacyAPIResponse
from openai.types.beta.threads.run import Run
from opentelemetry import context as context_api
from opentelemetry.context import _SUPPRESS_INSTRUMENTATION_KEY
from opentelemetry.semconv.ai import LLMRequestTypeValues, SpanAttributes
from opentelemetry.trace import SpanKind

from ..shared import _set_span_attribute, model_as_dict
from ..shared.config import Config
from ..utils import _with_tracer_wrapper, dont_throw

logger = logging.getLogger(__name__)

assistants = {}
runs = {}


@_with_tracer_wrapper
def assistants_create_wrapper(tracer, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    response = wrapped(*args, **kwargs)

    assistants[response.id] = {
        "model": kwargs.get("model"),
        "instructions": kwargs.get("instructions"),
    }

    return response


@_with_tracer_wrapper
def runs_create_wrapper(tracer, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    thread_id = kwargs.get("thread_id")
    instructions = kwargs.get("instructions")

    response = wrapped(*args, **kwargs)

    runs[thread_id] = {
        "start_time": time.time_ns(),
        "assistant_id": kwargs.get("assistant_id"),
        "instructions": instructions,
    }

    return response


@_with_tracer_wrapper
def runs_retrieve_wrapper(tracer, wrapped, instance, args, kwargs):
    @dont_throw
    def process_response(response):
        if type(response) is LegacyAPIResponse:
            parsed_response = response.parse()
        else:
            parsed_response = response
        assert type(parsed_response) is Run

        if parsed_response.id in runs:
            runs[thread_id]["end_time"] = time.time_ns()

    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    thread_id = kwargs.get("thread_id")
    response = wrapped(*args, **kwargs)
    process_response(response)

    return response


@_with_tracer_wrapper
def messages_list_wrapper(tracer, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    id = kwargs.get("thread_id")

    response = wrapped(*args, **kwargs)

    response_dict = model_as_dict(response)
    if id not in runs:
        return response

    run = runs[id]
    messages = sorted(response_dict["data"], key=lambda x: x["created_at"])

    span = tracer.start_span(
        "openai.assistant.run",
        kind=SpanKind.CLIENT,
        attributes={SpanAttributes.LLM_REQUEST_TYPE: LLMRequestTypeValues.CHAT.value},
        start_time=run.get("start_time"),
    )

    i = 0
    if assistants.get(run["assistant_id"]) is not None or Config.enrich_assistant:
        if Config.enrich_assistant:
            assistant = model_as_dict(
                instance._client.beta.assistants.retrieve(run["assistant_id"])
            )
            assistants[run["assistant_id"]] = assistant
        else:
            assistant = assistants[run["assistant_id"]]

        _set_span_attribute(
            span,
            SpanAttributes.LLM_REQUEST_MODEL,
            assistant["model"],
        )
        _set_span_attribute(
            span,
            SpanAttributes.LLM_RESPONSE_MODEL,
            assistant["model"],
        )
        _set_span_attribute(span, f"{SpanAttributes.LLM_PROMPTS}.{i}.role", "system")
        _set_span_attribute(
            span,
            f"{SpanAttributes.LLM_PROMPTS}.{i}.content",
            assistant["instructions"],
        )
        i += 1
    _set_span_attribute(span, f"{SpanAttributes.LLM_PROMPTS}.{i}.role", "system")
    _set_span_attribute(
        span, f"{SpanAttributes.LLM_PROMPTS}.{i}.content", run["instructions"]
    )

    for i, msg in enumerate(messages):
        prefix = f"{SpanAttributes.LLM_COMPLETIONS}.{i}"
        content = msg.get("content")

        _set_span_attribute(span, f"{prefix}.role", msg.get("role"))
        _set_span_attribute(
            span, f"{prefix}.content", content[0].get("text").get("value")
        )

    span.end(run.get("end_time"))

    return response


@_with_tracer_wrapper
def runs_create_and_stream_wrapper(tracer, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        return wrapped(*args, **kwargs)

    assistant_id = kwargs.get("assistant_id")
    instructions = kwargs.get("instructions")

    span = tracer.start_span(
        "openai.assistant.run_stream",
        kind=SpanKind.CLIENT,
        attributes={SpanAttributes.LLM_REQUEST_TYPE: LLMRequestTypeValues.CHAT.value},
    )

    i = 0
    if assistants.get(assistant_id) is not None or Config.enrich_assistant:
        if Config.enrich_assistant:
            assistant = model_as_dict(
                instance._client.beta.assistants.retrieve(assistant_id)
            )
            assistants[assistant_id] = assistant
        else:
            assistant = assistants[assistant_id]

        _set_span_attribute(
            span, SpanAttributes.LLM_REQUEST_MODEL, assistants[assistant_id]["model"]
        )
        _set_span_attribute(
            span,
            SpanAttributes.LLM_RESPONSE_MODEL,
            assistants[assistant_id]["model"],
        )
        _set_span_attribute(span, f"{SpanAttributes.LLM_PROMPTS}.{i}.role", "system")
        _set_span_attribute(
            span,
            f"{SpanAttributes.LLM_PROMPTS}.{i}.content",
            assistants[assistant_id]["instructions"],
        )
        i += 1
    _set_span_attribute(span, f"{SpanAttributes.LLM_PROMPTS}.{i}.role", "system")
    _set_span_attribute(span, f"{SpanAttributes.LLM_PROMPTS}.{i}.content", instructions)

    from .event_handler_wrapper import (
        EventHandleWrapper,
    )

    kwargs["event_handler"] = EventHandleWrapper(
        original_handler=kwargs["event_handler"], span=span
    )

    response = wrapped(*args, **kwargs)

    return response



import logging
from typing import Optional, Union

from .config import IudexConfig, _IudexConfig
from .utils import maybe_instrument_lib

logger = logging.getLogger(__name__)

INSTRUMENTATION_LIBS = [
    # iudex custom
    (
        ".openai",
        "OpenAIInstrumentor",
        {
            "enrich_assistant": True,
            "enrich_token_usage": True,
            "exception_logger": logger.error,
        },
    ),
    (".supabase", "SupabaseInstrumentor", {}),

    # opentelemetry-python-contrib
    # NOTE: these are excluded because they are already instrumented by higher level libs
    # - asgi
    # - dbapi
    # - starlette
    # - wsgi
    ("opentelemetry.instrumentation.aio_pika", "AioPikaInstrumentor", {}),
    ("opentelemetry.instrumentation.aiohttp_client", "AioHttpClientInstrumentor", {}),
    # broken: https://github.com/open-telemetry/opentelemetry-python-contrib/issues/2053
    # ("opentelemetry.instrumentation.aiohttp_server", "AioHttpServerInstrumentor", {}),
    ("opentelemetry.instrumentation.aiopg", "AiopgInstrumentor", {}),
    ("opentelemetry.instrumentation.asyncio", "AsyncioInstrumentor", {}),
    ("opentelemetry.instrumentation.asyncpg", "AsyncPGInstrumentor", {}),
    ("opentelemetry.instrumentation.aws_lambda", "AwsLambdaInstrumentor", {}),
    ("opentelemetry.instrumentation.boto", "BotoInstrumentor", {}),
    ("opentelemetry.instrumentation.boto3sqs", "Boto3SQSInstrumentor", {}),
    ("opentelemetry.instrumentation.botocore", "BotocoreInstrumentor", {}),
    ("opentelemetry.instrumentation.cassandra", "CassandraInstrumentor", {}),
    ("opentelemetry.instrumentation.celery", "CeleryInstrumentor", {}),
    ("opentelemetry.instrumentation.confluent_kafka", "ConfluentKafkaInstrumentor", {}),
    ("opentelemetry.instrumentation.django", "DjangoInstrumentor", {}),
    ("opentelemetry.instrumentation.elasticsearch", "ElasticsearchInstrumentor", {}),
    ("opentelemetry.instrumentation.falcon", "FalconInstrumentor", {}),
    ("opentelemetry.instrumentation.fastapi", "FastAPIInstrumentor", {}),
    ("opentelemetry.instrumentation.flask", "FlaskInstrumentor", {}),
    ("opentelemetry.instrumentation.grpc", ["GrpcInstrumentorClient", "GrpcInstrumentorServer"], {}),
    ("opentelemetry.instrumentation.httpx", "HTTPXClientInstrumentor", {}),
    ("opentelemetry.instrumentation.jinja2", "Jinja2Instrumentor", {}),
    ("opentelemetry.instrumentation.kafka_python", "KafkaInstrumentor", {}),
    ("opentelemetry.instrumentation.logging", "LoggingInstrumentor", {}),
    ("opentelemetry.instrumentation.mysql", "MySQLInstrumentor", {}),
    ("opentelemetry.instrumentation.mysqlclient", "MySQLClientInstrumentor", {}),
    ("opentelemetry.instrumentation.pika", "PikaInstrumentor", {}),
    ("opentelemetry.instrumentation.psycopg", "PsycopgInstrumentor", {}),
    ("opentelemetry.instrumentation.psycopg2", "Psycopg2Instrumentor", {}),
    ("opentelemetry.instrumentation.pymemcache", "PymemcacheInstrumentor", {}),
    ("opentelemetry.instrumentation.pymongo", "PymongoInstrumentor", {}),
    ("opentelemetry.instrumentation.pymysql", "PyMySQLInstrumentor", {}),
    ("opentelemetry.instrumentation.pyramid", "PyramidInstrumentor", {}),
    ("opentelemetry.instrumentation.redis", "RedisInstrumentor", {}),
    ("opentelemetry.instrumentation.remoulade", "RemouladeInstrumentor", {}),
    ("opentelemetry.instrumentation.requests", "RequestsInstrumentor", {}),
    (
        "opentelemetry.instrumentation.sqlalchemy",
        "SQLAlchemyInstrumentor",
        {"enable_commenter": True},
    ),
    ("opentelemetry.instrumentation.sqlite3", "SQLite3Instrumentor", {}),
    ("opentelemetry.instrumentation.system_metrics", "SystemMetricsInstrumentor", {}),
    ("opentelemetry.instrumentation.threading", "ThreadingInstrumentor", {}),
    ("opentelemetry.instrumentation.tornado", "TornadoInstrumentor", {}),
    ("opentelemetry.instrumentation.tortoiseorm", "TortoiseORMInstrumentor", {}),
    ("opentelemetry.instrumentation.urllib", "URLLibInstrumentor", {}),
    ("opentelemetry.instrumentation.urllib3", "URLLib3Instrumentor", {}),

    # openllmetry
    ("opentelemetry.instrumentation.alephalpha", "AlephAlphaInstrumentor", {}),
    ("opentelemetry.instrumentation.anthropic", "AnthropicInstrumentor", {}),
    ("opentelemetry.instrumentation.bedrock", "BedrockInstrumentor", {}),
    ("opentelemetry.instrumentation.chromadb", "ChromaInstrumentor", {}),
    ("opentelemetry.instrumentation.cohere", "CohereInstrumentor", {}),
    ("opentelemetry.instrumentation.google_generativeai", "GoogleGenerativeAiInstrumentor", {}),
    ("opentelemetry.instrumentation.haystack", "HaystackInstrumentor", {}),
    ("opentelemetry.instrumentation.langchain", "LangchainInstrumentor", {}),
    ("opentelemetry.instrumentation.llamaindex", "LlamaIndexInstrumentor", {}),
    ("opentelemetry.instrumentation.milvus", "MilvusInstrumentor", {}),
    ("opentelemetry.instrumentation.mistralai", "MistralAiInstrumentor", {}),
    ("opentelemetry.instrumentation.ollama", "OllamaInstrumentor", {}),
    ("opentelemetry.instrumentation.pinecone", "PineconeInstrumentor", {}),
    ("opentelemetry.instrumentation.qdrant", "QdrantInstrumentor", {}),
    ("opentelemetry.instrumentation.replicate", "ReplicateInstrumentor", {}),
    ("opentelemetry.instrumentation.together", "TogetherAiInstrumentor", {}),
    ("opentelemetry.instrumentation.transformers", "TransformersInstrumentor", {}),
    ("opentelemetry.instrumentation.vertexai", "VertexAIInstrumentor", {}),
    ("opentelemetry.instrumentation.watsonx", "WatsonxInstrumentor", {}),
    ("opentelemetry.instrumentation.weaviate", "WeaviateInstrumentor", {})
]


def instrument(
    service_name: Optional[str] = None,
    instance_id: Optional[str] = None,
    iudex_api_key: Optional[str] = None,
    log_level: Optional[Union[int, str]] = None,
    git_commit: Optional[str] = None,
    github_url: Optional[str] = None,
    env: Optional[str] = None,
    disable_print: Optional[bool] = None,
    config: Optional[IudexConfig] = None,
):
    """Auto-instruments app to send OTel signals to Iudex.

    Invoke this function in your app entrypoint.

    Args:
        service_name: Name of the service, e.g. "billing_service", __name__.
            If not supplied, env var OTEL_SERVICE_NAME will be used.
        instance_id: ID of the service instance, e.g. container id, pod name.
        iudex_api_key: Your Iudex API key.
            If not supplied, env var IUDEX_API_KEY will be used.
        log_level: Logging level for the root logger.
        git_commit: Commit hash of the currently deployed code.
            Used with github_url to deep link telemetry to source code.
        github_url: URL of the GitHub repository.
            Used with git_commit to deep link telemetry to source code.
        env: Environment of the service, e.g. "production", "staging".
        config: IudexConfig object with more granular options.
            Will override all other args, so provide them to the object instead.
    """
    for module_path, instrumentor_class_name, kwargs in INSTRUMENTATION_LIBS:
        if isinstance(instrumentor_class_name, list):
            for name in instrumentor_class_name:
                maybe_instrument_lib(module_path, name, **kwargs)
            continue
        maybe_instrument_lib(module_path, instrumentor_class_name, **kwargs)

    config = config or {
        "iudex_api_key": iudex_api_key,
        "service_name": service_name,
        "instance_id": instance_id,
        "logs_endpoint": None,
        "traces_endpoint": None,
        "log_level": log_level,
        "git_commit": git_commit,
        "github_url": github_url,
        "disable_print": disable_print,
        "env": env,
    }
    iudex_config = _IudexConfig(**config)
    iudex_config.configure()

    return iudex_config



from typing import Collection

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.metrics import get_meter
from opentelemetry.semconv.ai import Meters
from opentelemetry.trace import get_tracer
from wrapt import wrap_function_wrapper

from ..shared.chat_wrappers import achat_wrapper, chat_wrapper
from ..shared.completion_wrappers import acompletion_wrapper, completion_wrapper
from ..shared.embeddings_wrappers import aembeddings_wrapper, embeddings_wrapper
from ..utils import is_metrics_enabled
from ..version import __version__

_instruments = ("openai >= 0.27.0", "openai < 1.0.0")


class OpenAIV0Instrumentor(BaseInstrumentor):
    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, __version__, tracer_provider)

        meter_provider = kwargs.get("meter_provider")
        meter = get_meter(__name__, __version__, meter_provider)

        if is_metrics_enabled():
            tokens_histogram = meter.create_histogram(
                name=Meters.LLM_TOKEN_USAGE,
                unit="token",
                description="Measures number of input and output tokens used",
            )

            chat_choice_counter = meter.create_counter(
                name=Meters.LLM_GENERATION_CHOICES,
                unit="choice",
                description="Number of choices returned by chat completions call",
            )

            duration_histogram = meter.create_histogram(
                name=Meters.LLM_OPERATION_DURATION,
                unit="s",
                description="GenAI operation duration",
            )

            chat_exception_counter = meter.create_counter(
                name=Meters.LLM_COMPLETIONS_EXCEPTIONS,
                unit="time",
                description="Number of exceptions occurred during chat completions",
            )

            streaming_time_to_first_token = meter.create_histogram(
                name=Meters.LLM_STREAMING_TIME_TO_FIRST_TOKEN,
                unit="s",
                description="Time to first token in streaming chat completions",
            )
            streaming_time_to_generate = meter.create_histogram(
                name=Meters.LLM_STREAMING_TIME_TO_GENERATE,
                unit="s",
                description="Time between first token and completion in streaming chat completions",
            )
        else:
            (
                tokens_histogram,
                chat_choice_counter,
                duration_histogram,
                chat_exception_counter,
                streaming_time_to_first_token,
                streaming_time_to_generate,
            ) = (None, None, None, None, None, None)

        if is_metrics_enabled():
            embeddings_vector_size_counter = meter.create_counter(
                name=Meters.LLM_EMBEDDINGS_VECTOR_SIZE,
                unit="element",
                description="he size of returned vector",
            )
            embeddings_exception_counter = meter.create_counter(
                name=Meters.LLM_EMBEDDINGS_EXCEPTIONS,
                unit="time",
                description="Number of exceptions occurred during embeddings operation",
            )
        else:
            (
                tokens_histogram,
                embeddings_vector_size_counter,
                embeddings_exception_counter,
            ) = (None, None, None)

        wrap_function_wrapper("openai", "Completion.create", completion_wrapper(tracer))
        wrap_function_wrapper(
            "openai", "Completion.acreate", acompletion_wrapper(tracer)
        )
        wrap_function_wrapper(
            "openai",
            "ChatCompletion.create",
            chat_wrapper(
                tracer,
                tokens_histogram,
                chat_choice_counter,
                duration_histogram,
                chat_exception_counter,
                streaming_time_to_first_token,
                streaming_time_to_generate,
            ),
        )
        wrap_function_wrapper(
            "openai",
            "ChatCompletion.acreate",
            achat_wrapper(
                tracer,
                tokens_histogram,
                chat_choice_counter,
                duration_histogram,
                chat_exception_counter,
                streaming_time_to_first_token,
                streaming_time_to_generate,
            ),
        )
        wrap_function_wrapper(
            "openai",
            "Embedding.create",
            embeddings_wrapper(
                tracer,
                tokens_histogram,
                embeddings_vector_size_counter,
                duration_histogram,
                embeddings_exception_counter,
            ),
        )
        wrap_function_wrapper(
            "openai",
            "Embedding.acreate",
            aembeddings_wrapper(
                tracer,
                tokens_histogram,
                embeddings_vector_size_counter,
                duration_histogram,
                embeddings_exception_counter,
            ),
        )

    def _uninstrument(self, **kwargs):
        pass
